// Copyright 2000-2023 JetBrains s.r.o. and contributors. Use of this source code is governed by the Apache 2.0 license.
package com.intellij.openapi.util.text;

import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.*;

import java.awt.event.KeyEvent;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * An immutable object which represents a text string with mnemonic character.
 */
public final class TextWithMnemonic {
  public static final TextWithMnemonic EMPTY = new TextWithMnemonic("", -1, "");
  public static final Pattern MNEMONIC = Pattern.compile(" ?\\(_?[A-Z]\\)");

  private final @NotNull @Nls String text;
  /**
   * Mnemonic index (-1 = no mnemonic)
   */
  private final int mnemonicIndex;
  /**
   * A text that can be appended to myText to display a mnemonic that doesn't belong to the text naturally.
   * Ex: "Help (P)" - " (P)" will be extracted into a suffix.
   */
  private final @NotNull @Nls String mnemonicSuffix;

  private TextWithMnemonic(@NotNull @Nls String text, int mnemonicIndex, @NotNull @Nls String mnemonicSuffix) {
    assert mnemonicIndex >= 0 || mnemonicSuffix.isEmpty();
    assert mnemonicIndex >= -1 && mnemonicIndex < text.length() + mnemonicSuffix.length();
    this.text = text;
    this.mnemonicIndex = mnemonicIndex;
    this.mnemonicSuffix = mnemonicSuffix;
  }

  /**
   * @return plain text without mnemonic
   */
  public @NotNull @Nls String getText() {
    return getText(false);
  }

  /**
   * @param withMnemonicSuffix if true, add the mnemonic suffix (but without mnemonic)
   * @return plain text without mnemonic
   */
  public @NotNull @Nls String getText(boolean withMnemonicSuffix) {
    if (mnemonicSuffix.isEmpty()) {
      return text;
    }
    if (withMnemonicSuffix) {
      return text + mnemonicSuffix;
    }
    int ellipsisLength = getEllipsisLength(mnemonicSuffix);
    if (ellipsisLength == 0) {
      return text;
    }
    return text + mnemonicSuffix.substring(mnemonicSuffix.length() - ellipsisLength);
  }

  /**
   * @return a mnemonic character (upper-cased) if mnemonic is set; 0 otherwise
   * @deprecated use {@link #getMnemonicChar} or {@link #getMnemonicCode} instead
   */
  @Deprecated(forRemoval = true)
  public int getMnemonic() {
    char ch = getMnemonicChar();
    return ch == KeyEvent.CHAR_UNDEFINED ? 0 : Character.toUpperCase(ch);
  }

  /**
   * @return an extended key code for a mnemonic character, or {@link KeyEvent#VK_UNDEFINED} if mnemonic is not set
   */
  public int getMnemonicCode() {
    char ch = getMnemonicChar();
    return ch == KeyEvent.CHAR_UNDEFINED ? KeyEvent.VK_UNDEFINED : KeyEvent.getExtendedKeyCodeForChar(ch);
  }

  /**
   * @return a mnemonic character, or {@link KeyEvent#CHAR_UNDEFINED} if mnemonic is not set
   */
  public char getMnemonicChar() {
    if (mnemonicIndex < 0) return KeyEvent.CHAR_UNDEFINED;
    int index = mnemonicIndex - text.length();
    return index < 0 ? text.charAt(mnemonicIndex) : mnemonicSuffix.charAt(index);
  }

  /**
   * @return a mnemonic index if it's set; -1 otherwise
   */
  public int getMnemonicIndex() {
    return mnemonicIndex;
  }

  /**
   * @return true if mnemonic is set
   */
  public boolean hasMnemonic() {
    return mnemonicIndex >= 0;
  }

  /**
   * Drops a mnemonic
   * @return a TextWithMnemonic object where mnemonic is not set
   */
  public TextWithMnemonic dropMnemonic(boolean forceRemove) {
    if (!hasMnemonic()) return this;
    if (!forceRemove) return fromPlainText(text);

    Matcher matcher = MNEMONIC.matcher(text);
    if (matcher.find()) {
      //noinspection HardCodedStringLiteral
      return fromPlainText(matcher.replaceAll(""));
    }
    return this;
  }

  /**
   * Sets mnemonic at given index
   * @param index index, must be within the {@link #getText() text} string.
   * @return a TextWithMnemonic object with a mnemonic set at given index
   * @deprecated use {@link #withMnemonicIndex} or {@link #fromPlainTextWithIndex(String, int)} instead
   */
  @Deprecated(forRemoval = true)
  public TextWithMnemonic setMnemonicAt(int index) {
    if (index < 0 || index >= text.length() + mnemonicSuffix.length()) {
      throw new IndexOutOfBoundsException(String.valueOf(index));
    }
    return index == mnemonicIndex ? this : new TextWithMnemonic(text, index, mnemonicSuffix);
  }

  /**
   * @param index mnemonic index within the {@link #getText() text}, or {@code -1} to remove mnemonic
   * @return new {@code TextWithMnemonic} object with updated mnemonic index, or the same instance otherwise
   * @throws IndexOutOfBoundsException if mnemonic index cannot be used
   */
  public @NotNull TextWithMnemonic withMnemonicIndex(int index) {
    if (index == mnemonicIndex) return this;
    if (index >= -1) {
      String text = getText();
      if (index == -1) return fromPlainText(text);
      if (index < text.length() - getEllipsisLength(text)) return fromPlainTextWithIndex(text, index);
    }
    throw new IndexOutOfBoundsException(String.valueOf(index));
  }

  /**
   * Appends given a text to the current text.
   *
   * @param textToAppend text to append. Appended text is treated as a plain text, without mnemonic, so mnemonic position is unchanged.
   * @return TextWithMnemonic object which text is the concatenation of this object text and supplied text.
   */
  public TextWithMnemonic append(@NotNull @Nls String textToAppend) {
    return new TextWithMnemonic(text + textToAppend,
                                mnemonicIndex < text.length() ? mnemonicIndex : mnemonicIndex + textToAppend.length(),
                                mnemonicSuffix);
  }

  /**
   * Replaces the first occurrence of a given target text with the given replacement text.
   *
   * @param target the target text to be replaced
   * @param replacement the replacement text which is treated as a plain text, without mnemonic.
   * @return TextWithMnemonic object. The resulting mnemonic position could be adjusted if the mnemonic was located after the replacement.
   *          If the mnemonic was inside the target text, then it's dropped. Return this object if the target text was not found.
   */
  public TextWithMnemonic replaceFirst(@NotNull String target, @Nls @NotNull String replacement) {
    int index = text.indexOf(target);
    if (index == -1) {
      return this;
    }
    String resultText = text.substring(0, index) + replacement + text.substring(index + target.length());
    int resultIndex = mnemonicIndex < index ? mnemonicIndex :
                      mnemonicIndex >= index + target.length() ? mnemonicIndex - target.length() + replacement.length() :
                      -1;
    return new TextWithMnemonic(resultText, resultIndex, mnemonicSuffix);
  }

  /**
   * @param text a text with a mnemonic specified by the {@link UIUtil#MNEMONIC MNEMONIC} marker
   * @return new {@code TextWithMnemonic} object, or {@code null} if mnemonic is not specified in the given text
   * @throws IllegalArgumentException if the given text contains a marker at wrong position, or if it contains several markers
   * @see UIUtil#replaceMnemonicAmpersand
   */
  @ApiStatus.Internal
  public static @Nullable TextWithMnemonic fromMnemonicText(@NotNull @Nls String text) {
    int pos = text.indexOf(UIUtil.MNEMONIC);
    if (pos < 0) return null;
    String str = text.substring(pos + 1);
    if (str.isEmpty()) throw new IllegalArgumentException("unexpected mnemonic marker in " + text);
    if (str.indexOf(UIUtil.MNEMONIC) >= 0) throw new IllegalArgumentException("several mnemonic markers in " + text);
    return fromPlainTextWithIndex(pos > 0 ? text.substring(0, pos) + str : str, pos);
  }

  /**
   * Creates a TextWithMnemonic object from a plain text without mnemonic.
   * @param text a plain text to create a TextWithMnemonic object from
   * @return new TextWithMnemonic object which has no mnemonic
   */
  @Contract(pure = true)
  public static @NotNull TextWithMnemonic fromPlainText(@NotNull @Nls String text) {
    return text.isEmpty() ? EMPTY : new TextWithMnemonic(text, -1, "");
  }

  /**
   * Creates a TextWithMnemonic object from a plain text without mnemonic.
   * @param text a plain text to create a TextWithMnemonic object from
   * @param mnemonicChar mnemonic character (0 = absent mnemonic)
   * @return new TextWithMnemonic object which has given mnemonic character.
   * If the text doesn't contain the supplied character, then mnemonicChar is appended in parentheses.
   */
  @Contract(pure = true)
  public static @NotNull TextWithMnemonic fromPlainText(@NotNull @Nls String text, char mnemonicChar) {
    if (mnemonicChar == 0) {
      return fromPlainText(text);
    }
    mnemonicChar = Character.toUpperCase(mnemonicChar);
    for (int i = 0; i < text.length(); i++) {
      if (Character.toUpperCase(text.charAt(i)) == mnemonicChar) {
        return new TextWithMnemonic(text, i, "");
      }
    }
    int ellipsisLength = getEllipsisLength(text);
    String suffix = "(" + mnemonicChar + ")" + text.substring(text.length() - ellipsisLength);
    text = text.substring(0, text.length() - ellipsisLength);
    return new TextWithMnemonic(text, text.length() + 1, suffix);
  }

  /**
   * Creates new {@code TextWithMnemonic} object from a plain text with a mnemonic specified by its index.
   *
   * @param text  a plain text
   * @param index a mnemonic index in the given text, or {@code -1} if it is not needed
   * @return new {@code TextWithMnemonic} object that may have a mnemonic
   */
  public static @NotNull TextWithMnemonic fromPlainTextWithIndex(@NotNull @Nls String text, int index) {
    if (index < 0) return fromPlainText(text);
    if (index < text.length()) {
      // try to extract a mnemonic suffix
      int pos = text.length() - getEllipsisLength(text) - 3; // the length of "(M)"
      if (pos >= 0 && index == (pos + 1) && text.charAt(pos) == '(' && text.charAt(pos + 2) == ')') {
        while (pos > 0 && text.charAt(pos - 1) == ' ') pos--; // skip spaces
        return new TextWithMnemonic(text.substring(0, pos), index, text.substring(pos));
      }
    }
    return new TextWithMnemonic(text, index, "");
  }

  /**
   * Parses a text in text-with-mnemonic format.
   * A mnemonic is prepended either with '_', or with '&' or with '\x1B' character.
   * To escape '_' or '&' before the actual mnemonic the character must be duplicated.
   * The characters after the actual mnemonic should not be escaped.
   * E.g. "A__b_c__d" in text-with-mnemonic format will be displayed as "A_bc__d" with mnemonic 'c'.
   *
   * @param text text to parse
   * @return TextWithMnemonic object which corresponds to the parsed text.
   */
  @Contract(pure = true)
  public static @NotNull TextWithMnemonic parse(@NotNull @Nls String text) {
    TextWithMnemonic mnemonic = text.isEmpty() ? EMPTY : fromMnemonicText(text);
    if (mnemonic != null) {
      return mnemonic;
    }

    if (text.contains("_") || text.contains("&")) {
      @Nls StringBuilder plainText = new StringBuilder();
      int mnemonicIndex = -1;

      int backShift = 0;
      for (int i = 0; i < text.length(); i++) {
        char ch = text.charAt(i);
        if (mnemonicIndex == -1 && (ch == '_' || ch == '&')) {
          //noinspection AssignmentToForLoopParameter
          i++;
          if (i >= text.length()) break;
          ch = text.charAt(i);
          if (ch != '_' && ch != '&') {
            mnemonicIndex = i - 1 - backShift;
          }
          else {
            backShift++;
          }
        }
        plainText.append(ch);
      }
      return fromPlainTextWithIndex(plainText.toString(), mnemonicIndex);
    }
    return fromPlainText(text);
  }

  private static int getEllipsisLength(String text) {
    if (text.endsWith("...")) {
      return 3;
    }
    if (text.endsWith("…")) {
      return 1;
    }
    return 0;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (o == null || getClass() != o.getClass()) return false;
    TextWithMnemonic mnemonic = (TextWithMnemonic)o;
    return mnemonicIndex == mnemonic.mnemonicIndex &&
           text.equals(mnemonic.text) &&
           mnemonicSuffix.equals(mnemonic.mnemonicSuffix);
  }

  @Override
  public int hashCode() {
    return (text.hashCode() * 31 + mnemonicIndex) * 31 + mnemonicSuffix.hashCode();
  }

  /**
   * @return text in text-with-mnemonic format. Parsing back this text using {@link #parse(String)} method would create
   * a TextWithMnemonic object which is equal to this.
   */
  @Override
  public @Nls String toString() {
    if (mnemonicIndex > -1) {
      String completeText = text + mnemonicSuffix;
      String prefix = StringUtil.escapeMnemonics(completeText.substring(0, mnemonicIndex));
      String suffix = completeText.substring(mnemonicIndex);
      return prefix + "_" + suffix;
    }
    return StringUtil.escapeMnemonics(text);
  }
}
